#include <cstdlib>
using std::exit;

#include <cstddef>
using std::size_t;

#include <sstream>
using std::stringstream;

#include <cassert>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
using std::fixed;

#include <fstream>
using std::ofstream;

#include <iomanip>
using std::setw;

#include <cmath>

#include <vector>
using std::vector;

#include <algorithm>
using std::sort;

#include "particle.hpp"

vector<particle> gas;
const double dt = 0.0005;
const double dt2 = dt*dt;
double global_time = 0.0; // !! a global time counter

vector<double> radial_dist(1001); // !! data structure for histogram of the radial distribution function 
size_t radial_count;

vector< vector<double> > velocity_dist; // !! data structure for histogram of the velocity distribution function 
vector<size_t> velocity_count(5);

const int delay = 1; // milliseconds

void glutPrint(float x, float y, const char* text, float r, float g, float b, float a) 
{ 
    if(!text || !strlen(text)) return; 
    bool blending = false; 
    if(glIsEnabled(GL_BLEND)) blending = true; 
    glEnable(GL_BLEND); 
    glColor4f(r,g,b,a); 
    glRasterPos2f(x,y); 
    while (*text) { 
        glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *text); 
        text++; 
    } 
    if(!blending) glDisable(GL_BLEND); 
}

void display()
{
	glClear(GL_COLOR_BUFFER_BIT);
	for (size_t n = 0; n < gas.size(); ++n)
		gas[n].draw();
	stringstream s;
	s.precision(1);
	s << "t = " << fixed << global_time;
	glutPrint(-0.85f, -0.85f,s.str().c_str(), 1.0f, 1.0f, 1.0f, 0.5f);
	glFlush();
	glutSwapBuffers();
}

void keyboard(unsigned char key, int x, int y)
{
	if (key == 'q')
		exit(0);
}

#include "mtrand.hpp"
mtrand Rand;
inline double R(void) { return 2.0*Rand()-1.0; } // !! generates random numbers in [-1,1]

const double dx = 0.01;
#ifdef USE_GAS_INITIAL_COND
const double dv = 4.0;
#endif
#ifdef USE_SOLID_INITIAL_COND
const double dv = 0.0;
#endif
void initialize_gas(void)
{
	for (double x = -0.9; x < 1.0; x += 0.2)
		for (double y = -0.9; y < 1.0; y += 0.2)
			gas.push_back( particle( /*mass*/      1.0,
			                         /*position*/  vec2(x+dx*R(),y+dx*R()),
					                 /*velocity*/  vec2(dv*R(),dv*R()) ) );
	cout << "Simulating " << gas.size() << " particles..." << endl;
	velocity_dist.resize(5); // !! resize vectors for velocity histogram
	velocity_dist[0].resize(1001);
	velocity_dist[1].resize(1001);
	velocity_dist[2].resize(1001);
	velocity_dist[3].resize(1001);
	velocity_dist[4].resize(1001);
}

inline double map_back(double x)
{
	// !! map position x back into [-1,1]
	const double d = int(fabs(x));
	if (x > 1.0) return -1.0 + x - d;
	if (x < -1.0) return 1.0 + x + d;
	return x;
}

inline double diff(double x, double y)
{
	// !! redefinition of differences in the compact space
	const double d = x-y;
	if (d > 1.0) return d-2.0;
	if (d < -1.0) return 2.0+d;
	return d;
}

inline double angle(const particle &p1, const particle &p2)
{
	const double dx = diff(p2.r.x,p1.r.x);
	const double dy = diff(p2.r.y,p1.r.y);
	return atan2(dy,dx);
}

inline double distance2(const particle &p1, const particle &p2)
{
	const double dx = diff(p2.r.x,p1.r.x);
	const double dy = diff(p2.r.y,p1.r.y);
	return dx*dx + dy*dy;
}

const double epsilon = 1.0;
#ifdef USE_GAS_INITIAL_COND
const double sigma = 0.1;
#endif
#ifdef USE_SOLID_INITIAL_COND
const double sigma = 0.21;
#endif

const double sigma2 = sigma*sigma;
vec2 recompute_acceleration(size_t n)
{
	double sumx = 0.0;
	double sumy = 0.0;
	for (size_t m = 0; m < gas.size(); ++m)
		if (n != m)
		{
			// !! compute Lennard-Jones forces
			const double r2 = distance2(gas[n],gas[m]);
			const double r = sqrt(r2);
			const double theta = angle(gas[n],gas[m]);
			const double sigma2_r2 = sigma2/r2;
			const double F = -(24.0*epsilon/r)*( 2.0*pow(sigma2_r2,6.0)-pow(sigma2_r2,3.0) );
			sumx += cos(theta)*F;
			sumy += sin(theta)*F;
		}
	return vec2(sumx/gas[n].mass,sumy/gas[n].mass);
}

double potential_energy(void)
{
	double sum = 0.0;
	for (size_t m = 1; m < gas.size(); ++m)
		for (size_t n = 0; n < m; ++n)
		{
			const double r2 = distance2(gas[n],gas[m]);
			const double sigma2_r2 = sigma2/r2;
			sum += pow(sigma2_r2,6.0)-pow(sigma2_r2,3.0);
		}
	return 4.0*epsilon*sum;
}

double kinetic_energy(void)
{
	double sum = 0.0;
	for (size_t n = 0; n < gas.size(); ++n)
		sum += gas[n].mass*gas[n].v.norm2();
	return 0.5*sum;
}

void update_gas(void)
{
	// self-starting Verlet algorithm
	// x(n+1) = x(n) + v(n)dt + 0.5*a(n)dt^2
	// v(n+1) = v(n) + 0.5*(a(n+1)+a(n))*dt
	
	for (size_t n = 0; n < gas.size(); ++n)
		gas[n].a = recompute_acceleration(n); // a(n)
	for (size_t n = 0; n < gas.size(); ++n)
	{
		gas[n].r += gas[n].v*dt + gas[n].a*(0.5*dt2); // x(n) -> x(n+1)
		gas[n].r.x = map_back(gas[n].r.x);
		gas[n].r.y = map_back(gas[n].r.y);
	}
	for (size_t n = 0; n < gas.size(); ++n)
	{
		const vec2 a_old = gas[n].a;
		const vec2 a_new = gas[n].a = recompute_acceleration(n); // a(n+1)
		gas[n].v += (a_new+a_old)*0.5*dt;  // v(n+1) = v(n) + 0.5*(a(n+1)+a(n))*dt
	}
	global_time += dt; // !! increment the time counter
	
	if (global_time > 10.0)
	{
		// !! measure the radial distribution
		for (size_t m = 1; m < gas.size(); ++m)
			for (size_t n = 0; n < m; ++n)
			{
				const double d = sqrt(distance2(gas[n],gas[m]));
				++radial_dist[size_t(d*700)];
				++radial_count;
			}
	}

	// !! determine the measurement interval
	size_t k;
	bool is_measurement_interval = true;
	if (global_time > 0.1 and global_time < 0.25)
		k = 0;
	else if (global_time > 1.0 and global_time < 2.0)
		k = 1;	
	else if (global_time > 4.0 and global_time < 6.0)
		k = 2;
	else if (global_time > 9.0 and global_time < 12.0)
		k = 3;
	else if (global_time > 16.0 and global_time < 20.0)
		k = 4;
	else
		is_measurement_interval = false;

	if (is_measurement_interval)
	{
		// !! measure the velocity distribution
		assert(k < velocity_dist.size());
		assert(velocity_dist[k].size() == 1001);
		for (size_t n = 0; n < gas.size(); ++n)
		{
			const double v = sqrt(gas[n].v.norm2());
			const size_t i = size_t(v*40);
			if (i <= 1000) ++velocity_dist[k][i];
			++velocity_count[k];
		}
	}

	if (global_time > 20.0)
	{
		// !! write data to files and exit
#ifdef USE_GAS_INITIAL_COND
		ofstream fout("radial_dist_gas.dat");
#endif
#ifdef USE_SOLID_INITIAL_COND
		ofstream fout("radial_dist_solid.dat");
#endif
		for (size_t n = 0; n < 1001; ++n)
			fout << (1.0/700.0)*n << "\t" << (1400.0*radial_dist[n])/radial_count << endl;
		fout.close();
#ifdef USE_GAS_INITIAL_COND
		fout.open("velocity_dist_gas.dat");
#endif
#ifdef USE_SOLID_INITIAL_COND
		fout.open("velocity_dist_solid.dat");
#endif
		for (size_t k = 0; k < velocity_dist.size(); ++k)
		{
			for (size_t n = 0; n < 1001; ++n)
				fout << 0.025*n << "\t" << (40.0*velocity_dist[k][n])/velocity_count[k] << endl;
			fout << endl << endl;
		}
		exit(0);
	}
}

void animate(int)
{
	update_gas();
	glutPostRedisplay();
	glutTimerFunc(delay,animate,0);
}

int main(int argc, char **argv)
{
	glutInit(&argc,argv);
	glutInitDisplayMode (GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize (640, 640);
	glutInitWindowPosition (100, 100);
	glutCreateWindow("Dynamics of a monatomic gas");
	glutDisplayFunc(display);
   glutKeyboardFunc(keyboard);
	glutTimerFunc(delay,animate,0);
	initialize_gas();
	glutMainLoop();
}
