#include <cstdlib>
using std::exit;

#include <cstddef>
using std::size_t;

#include <cassert>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

#include <fstream>
using std::ofstream;

#include <iomanip>
using std::setw;

#include <cmath>

#include <vector>
using std::vector;

#include <algorithm>
using std::sort;

#include "particle.hpp"

vector<particle> gas; // gas = vector of particles
const double dt = 0.0005;
const double dt2 = dt*dt;
double global_time = 0.0;

const int delay = 1; // milliseconds

void display()
{
	glClear(GL_COLOR_BUFFER_BIT);
	for (size_t n = 0; n < gas.size(); ++n)
		gas[n].draw();
	glutSwapBuffers();
}

void keyboard(unsigned char key, int x, int y)
{
	if (key == 'q')
		exit(0);
}

#include "mtrand.hpp"
mtrand Rand;
inline double R(void) { return 2.0*Rand()-1.0; }

const double dx = 0.01;
#if (INIT_STYLE == 1)
const double dv = 4.0;
#elif (INIT_STYLE == 2)
const double dv = 0.25;
#endif

void initialize_gas(void)
{
	for (double x = -0.9; x < 1.0; x += 0.2)
		for (double y = -0.9; y < 1.0; y += 0.2)
			gas.push_back( particle( /*mass*/      1.0,
			                         /*position*/  vec2(x+dx*R(),y+dx*R()),
					                 /*velocity*/  vec2(dv*R(),dv*R()) ) );
	cout << "Simulating " << gas.size() << " particles..." << endl;
}

inline double map_back(double x)
{
	const double d = int(fabs(x));
	if (x > 1.0) return -1.0 + x - d;
	if (x < -1.0) return 1.0 + x + d;
	return x;
}

inline double diff(double x, double y)
{
	const double d = x-y;
	//assert(d <= 2.0);
	//assert(d >= -2.0);
	if (d > 1.0) return d-2.0;
	if (d < -1.0) return 2.0+d;
	return d;
}

inline double angle(const particle &p1, const particle &p2)
{
	const double dx = diff(p2.r.x,p1.r.x);
	const double dy = diff(p2.r.y,p1.r.y);
	return atan2(dy,dx);
}

inline double distance2(const particle &p1, const particle &p2)
{
	const double dx = diff(p2.r.x,p1.r.x);
	const double dy = diff(p2.r.y,p1.r.y);
	return dx*dx + dy*dy;
}

const double epsilon = 1.0;// 1.0;
#if (INIT_STYLE == 1)
const double sigma = 0.1; // 0.1; (gas) 0.21; (solid)
#elif (INIT_STYLE == 2)
const double sigma = 0.21; // 0.1; (gas) 0.21; (solid)
#endif
const double sigma2 = sigma*sigma;
vec2 recompute_acceleration(size_t n)
{
	double sumx = 0.0;
	double sumy = 0.0;
	for (size_t m = 0; m < gas.size(); ++m)
		if (n != m)
		{
			const double r2 = distance2(gas[n],gas[m]);
			const double r = sqrt(r2);
			const double theta = angle(gas[n],gas[m]);
			const double sigma2_r2 = sigma2/r2;
			const double F = -(24.0*epsilon/r)*( 2.0*pow(sigma2_r2,6.0)-pow(sigma2_r2,3.0) );
			sumx += cos(theta)*F;
			sumy += sin(theta)*F;
		}
	return vec2(sumx/gas[n].mass,sumy/gas[n].mass);
}

double potential_energy(void)
{
	double sum = 0.0;
	for (size_t m = 1; m < gas.size(); ++m)
		for (size_t n = 0; n < m; ++n)
		{
			const double r2 = distance2(gas[n],gas[m]);
			const double sigma2_r2 = sigma2/r2;
			sum += pow(sigma2_r2,6.0)-pow(sigma2_r2,3.0);
		}
	return 4.0*epsilon*sum;
}

double kinetic_energy(void)
{
	double sum = 0.0;
	for (size_t n = 0; n < gas.size(); ++n)
		sum += gas[n].mass*gas[n].v.norm2();
	return 0.5*sum;
}

void update_gas(void)
{
	// self-starting Verlet algorithm
	// x(n+1) = x(n) + v(n)dt + 0.5*a(n)dt^2
	// v(n+1) = v(n) + 0.5*(a(n+1)+a(n))*dt
	
	for (size_t n = 0; n < gas.size(); ++n)
		gas[n].a = recompute_acceleration(n); // a(n)
	for (size_t n = 0; n < gas.size(); ++n)
	{
		gas[n].r += gas[n].v*dt + gas[n].a*(0.5*dt2); // x(n) -> x(n+1)
		gas[n].r.x = map_back(gas[n].r.x);
		gas[n].r.y = map_back(gas[n].r.y);
	}
	for (size_t n = 0; n < gas.size(); ++n)
	{
		const vec2 a_old = gas[n].a;
		const vec2 a_new = gas[n].a = recompute_acceleration(n); // a(n+1)
		gas[n].v += (a_new+a_old)*0.5*dt;  // v(n+1) = v(n) + 0.5*(a(n+1)+a(n))*dt
	}
	global_time += dt;
}

void animate(int)
{
	update_gas();
	glutPostRedisplay();
	glutTimerFunc(delay,animate,0);
}

int main(int argc, char **argv)
{
	glutInit(&argc,argv);
	glutInitDisplayMode (GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize (640, 640);
	glutInitWindowPosition (100, 100);
	glutCreateWindow("Dynamics of a monatomic gas");
	glutDisplayFunc(display);
   glutKeyboardFunc(keyboard);   
	glutTimerFunc(delay,animate,0);
	initialize_gas();
	//fout.open("path.dat");
	glutMainLoop();
}
