#include <cstdlib>
using std::exit;

#include <cstddef>
using std::size_t;

#include <cassert>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

#include <fstream>
using std::ofstream;

#include <iomanip>
using std::setw;

#include <cmath>

#include <vector>
using std::vector;

#include "particle.hpp"

vector<particle> gas; // gas = vector of particles
const double dt = 0.0005;
const double dt2 = dt*dt;

const int delay = 1; // milliseconds

void display()
{
	glClear(GL_COLOR_BUFFER_BIT);
	for (size_t n = 0; n < gas.size(); ++n)
		gas[n].draw();
	glutSwapBuffers();
}

void initialize_gas(void)
{
	const vec2 x1( 0.48500218,-0.121543765);
	const vec2 v3(-1.31862315,-1.222914959);
	gas.push_back( particle( /*mass*/      1.0,
	                         /*position*/  x1,
	                         /*velocity*/  v3*(-0.5) ) );
	gas.push_back( particle( /*mass*/      1.0,
	                         /*position*/  x1*(-1.0),
	                         /*velocity*/  v3*(-0.5) ) );
	gas.push_back( particle( /*mass*/      1.0,
	                         /*position*/  vec2(0.0,0.0),
	                         /*velocity*/  v3 ) );
	cout << "Simulating " << gas.size() << " particles..." << endl;
}

inline double angle(const particle &p1, const particle &p2)
{
	const double dx = p2.r.x - p1.r.x;
	const double dy = p2.r.y - p1.r.y;
	return atan2(dy,dx);
}

inline double distance2(const particle &p1, const particle &p2)
{
	const double dx = p2.r.x - p1.r.x;
	const double dy = p2.r.y - p1.r.y;
	return dx*dx + dy*dy;
}

vec2 recompute_acceleration(size_t n)
{
	double sumx = 0.0;
	double sumy = 0.0;
	for (size_t m = 0; m < gas.size(); ++m)
		if (n != m)
		{
			const double r2 = distance2(gas[n],gas[m]);
			const double theta = angle(gas[n],gas[m]);
			sumx += gas[m].mass*cos(theta)/r2;
			sumy += gas[m].mass*sin(theta)/r2;
		}
	return vec2(sumx,sumy);
}

ofstream fout;
void update_gas(void)
{
	// self-starting Verlet algorithm
	// x(n+1) = x(n) + v(n)dt + 0.5*a(n)dt^2
	// v(n+1) = v(n) + 0.5*(a(n+1)+a(n))*dt
	
	for (size_t n = 0; n < gas.size(); ++n)
		gas[n].a = recompute_acceleration(n); // a(n)
	for (size_t n = 0; n < gas.size(); ++n)
		gas[n].r += gas[n].v*dt + gas[n].a*(0.5*dt2); // x(n) -> x(n+1)
	for (size_t n = 0; n < gas.size(); ++n)
	{
		const vec2 a_old = gas[n].a;
		const vec2 a_new = gas[n].a = recompute_acceleration(n); // a(n+1)
		gas[n].v += (a_new+a_old)*0.5*dt; // v(n+1) = v(n) + 0.5*(a(n+1)+a(n))*dt
	}
	gas[0].t += dt;
	fout << gas[0] << endl;
		if (gas[0].t > 20.0)
		{
			fout.close();
			exit(0);
		}
}

void animate(int)
{
	update_gas();
	glutPostRedisplay();
	glutTimerFunc(delay,animate,0);
}

int main(int argc, char **argv)
{
	glutInit(&argc,argv);
	glutInitDisplayMode (GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize (640, 640);
	glutInitWindowPosition (100, 100);
	glutCreateWindow("A three-planet solar system");
	glutDisplayFunc(display);
	glutTimerFunc(delay,animate,0);
	initialize_gas();
	fout.open("path.dat");
	glutMainLoop();
}
